import numpy as np
from jax.scipy.linalg import expm
import jax.numpy as jnp
from scipy.sparse import csr_matrix, csc_matrix

# from quantum_simulation_recipe.bounds import *
# from quantum_simulation_recipe.trotter import pf
# pf_r = pf

def commutator(A, B):
    return A @ B - B @ A

# def anticommutator(A, B, to_sparse=False):
def anticommutator(A, B):
    return A @ B + B @ A

def norm(A, ord='spectral'):
    if ord == 'fro':
        return np.linalg.norm(A)
    elif ord == 'spectral':
        return np.linalg.norm(A, ord=2)
    elif ord == '4':
        return np.trace(A @ A.conj().T @ A @ A.conj().T)**(1/4)
    else:
        # raise ValueError('norm is not defined')
        return np.linalg.norm(A, ord=ord)

# def measure_error(r, h_list, t, exact_U, type, rand_states=[], ob=None, pf_ord=2, coeffs=[], use_jax=False, verbose=False, return_error_list=False): 
#     # print(type)
#     if type == 'worst_empirical':
#         return 2 * np.linalg.norm(exact_U - pf_r(h_list, t, r, order=pf_ord, use_jax=use_jax), ord=2)
#     elif type == 'worst_bound':
#         if coeffs != []:
#             return 2 * tight_bound(h_list, 2, t, r) * coeffs[0]
#         else:
#             return 2 * tight_bound(h_list, 2, t, r)
#     elif type == 'worst_ob_empirical':
#         appro_U = pf_r(h_list, t, r, order=pf_ord, use_jax=use_jax)
#         # appro_U = pf_r(h_list, t, r, order=pf_ord)
#         exact_ob = exact_U.conj().T @ ob @ exact_U 
#         appro_ob = appro_U.conj().T @ ob @ appro_U
#         # ob_error = np.linalg.norm(exact_ob - appro_ob, ord=2)
#         ob_error = np.sort(abs(np.linalg.eigvalsh(exact_ob - appro_ob)))[-1]
#         print('ob error (operator norm, largest eigen): ', ob_error, '; r:', r, '; t:', t)
#         return ob_error
#     elif type == 'worst_loose_bound':
#         return relaxed_st_bound(r, coeffs[1], coeffs[2], t, ob_type=coeffs[0])
#     elif type == 'lightcone_bound':
#         return lc_tail_bound(r, coeffs[1], coeffs[2], t, ob_type=coeffs[0], verbose=False)
#         # return relaxed_lc_bound(r, coeffs[1], coeffs[2], t, ob_type=coeffs[0], verbose=False)
#     elif type == 'average_bound':
#         # return tight_bound(h_list, 2, t, r, type='4')
#         if coeffs != []:
#             return 2 * tight_bound(h_list, 2, t, r, type='fro') * coeffs[0]
#         else:
#             return 2 * tight_bound(h_list, 2, t, r, type='fro')
#     elif type == 'average_empirical':
#         appro_U = pf_r(h_list, t, r, order=pf_ord, use_jax=use_jax)
#         err_list = [np.linalg.norm(np.outer(exact_U @ state.data.conj().T , (exact_U @ state.data.conj().T).conj().T) - np.outer(appro_U @ state.data.conj().T, (appro_U @ state.data.conj().T).conj().T), ord='nuc') for state in rand_states]
#         # err_list = [np.linalg.norm((exact_U - pf_r(h_list, t, r, order=pf_ord, use_jax=use_jax)) @ state.data) for state in rand_states]
#         if type(ob) == list:
#             ob_norm = sum([np.linalg.norm(o, ord=2) for o in ob])
#         else:
#             ob_norm = np.linalg.norm(ob, ord=2)
#         if return_error_list:
#             return np.array(err_list) * ob_norm
#             # return np.array(err_list) * np.linalg.norm(ob, ord=2)
#         else:
#             # return np.mean(err_list) * np.linalg.norm(ob, ord=2)
#             return np.mean(err_list) * ob_norm
#     elif type == 'average_ob_bound_legacy':
#     # elif type == 'average_ob_bound':
#         if isinstance(h_list[0], csr_matrix):
#             onestep_exactU = scipy.linalg.expm(-1j * t/r * sum([herm.toarray() for herm in h_list]))
#             d = len(h_list[0].toarray())
#         elif isinstance(h_list[0], np.ndarray):
#             onestep_exactU = scipy.linalg.expm(-1j * t/r * sum([herm for herm in h_list]))
#             d = len(h_list[0])
#         E_op = onestep_exactU - pf_r(h_list, t/r, 1, order=pf_ord, use_jax=use_jax)
#         # print((np.trace(E_op @ E_op.conj().T @ E_op @ E_op.conj().T)/d)**(1/4))
#         bound = 2 * r * (np.trace(E_op @ E_op.conj().T @ E_op @ E_op.conj().T)/d)**(1/4) * (np.trace(ob @ ob @ ob @ ob)/d)**(1/4)
#         # print(f'bound_e={bound_e}, bound={bound}')
#         return bound
#     elif type == 'average_ob_bound':
#     # elif type == 'average_ob_bound_nc':
#         if isinstance(h_list[0], csr_matrix):
#             d = len(h_list[0].toarray())
#         elif isinstance(h_list[0], np.ndarray):
#             d = len(h_list[0])
#         # if coeffs == []:
#         #     bound = 2 * tight_bound(h_list, 2, t, r, type='4') * (np.trace(ob @ ob @ ob @ ob)/d)**(1/4)
#         # else:
#         #     bound = np.sqrt(2) * tight_bound(h_list, 2, t, r, type='fro') * (sum([np.linalg.norm(ob, ord='fro') for ob in coeffs[0]])/(d+1)**(1/2)) 
#         if type(ob) == list:
#             bound = np.sqrt(2) * tight_bound(h_list, 2, t, r, type='fro') * sum([np.linalg.norm(o, ord='fro') for o in ob])/(d+1)**(1/2) 
#         else:
#             bound = np.sqrt(2) * tight_bound(h_list, 2, t, r, type='fro') * np.linalg.norm(ob, ord='fro')/(d+1)**(1/2) 
#         return bound
#     # elif type == 'observable_empirical':
#     elif type == 'average_ob_empirical':
#         approx_U = pf_r(h_list, t, r, order=pf_ord, use_jax=use_jax)
#         exact_final_states = [exact_U @ state.data.T for state in rand_states]
#         appro_final_states = [approx_U @ state.data.T for state in rand_states]
#         if type(ob) == list:
#             err_list = [sum([abs(appro_final_states[i].conj().T @ o @ appro_final_states[i] - exact_final_states[i].conj().T @ o @ exact_final_states[i]) for o in ob]) for i in range(len(rand_states))]
#         else:
#             err_list = [abs(appro_final_states[i].conj().T @ ob @ appro_final_states[i] - exact_final_states[i].conj().T @ ob @ exact_final_states[i]) for i in range(len(rand_states))]
#         if return_error_list:
#             return np.array(err_list)
#         else:
#             return np.mean(err_list)
#     # elif type == 'observable_bound':
#     #     return None
#     else: 
#         raise ValueError(f'type={type} is not defined!')

# def op_error(exact, approx, norm='spectral'):
#     ''' 
#     Frobenius norm of the difference between the exact and approximated operator
#     input:
#         exact: exact operator
#         approx: approximated operator
#     return: error of the operator
#     '''
#     if norm == 'fro':
#         return jnp.linalg.norm(exact - approx)
#     elif norm == 'spectral':
#         # if the input is in csr_matrix format
#         if isinstance(exact, csc_matrix) and isinstance(approx, csc_matrix):
#             return jnp.linalg.norm(jnp.array(exact.toarray() - approx.toarray()), ord=2)
#         else:
#             return jnp.linalg.norm(exact - approx, ord=2)
#     else:
#         raise ValueError('norm is not defined')
#     # return np.linalg.norm(exact - approx)/len(exact)


# # evaluate trotter error for different number of trotter steps
# def trotter_error(list_herm, r_list, t, norm='spectral', n_perm=50, verbose=False):
#     ''' 
#     evaluate trotter error for different number of trotter steps
#     input: 
#         list_herm: a list of Hermitian matrices
#         r_list: number of trotter steps
#     return: trotter error
#     '''
#     exact_U = expm(-1j * t * sum(list_herm))
#     list_U = [expm(-1j * t / (2*r_list[-1]) * herm) for herm in list_herm]
#     if len(list_U) >= 5:
#         print('number of terms: ', len(list_U))
#         perm_list = [list_U] 
#         seed_value = random.randrange(sys.maxsize)
#         random.seed(seed_value)  
#         # randomly select 5 permutations from perm_list
#         for _ in range(n_perm-1):
#             # random.shuffle(list_U) 
#             # perm_list.append(list_U[:])
#             perm_list.append(random.sample(list_U, len(list_U)))
#         # perm_list = random.sample(perm_list, 50) 
#         print('# randomly selected perm: ', len(perm_list))
#     else:
#         # generate a list of permutation of the order of the matrices
#         perm_list = list(itertools.permutations(list_U))
#         # print('perm_list', perm_list)
#         print('# all perm: ', len(perm_list))
#     # perm_list = list(itertools.permutations(list_herm))[:5]
#     # for r in r_list:
#     # first-order trotter
#     trotter_error_list = [op_error(matrix_power(matrix_product(perm, int(2*r_list[-1]/r)), r), exact_U, norm) for r in r_list for perm in perm_list]
#     # trotter_error_list = [op_error(matrix_power(unitary_matrix_product(perm, t=t/r), r), exact_U, norm) for r in r_list for perm in perm_list]
#     # second-order trotter
#     trotter_error_list_2nd = [op_error(matrix_power(matrix_product(perm, int(r_list[-1]/r)) @ matrix_product(perm[::-1], int(r_list[-1]/r)), r), exact_U, norm) for r in r_list for perm in perm_list]
#     err_1st_reshaped = np.array(trotter_error_list).reshape(len(r_list), len(perm_list))
#     err_2nd_reshaped = np.array(trotter_error_list_2nd).reshape(len(r_list), len(perm_list))

#     return err_1st_reshaped , err_2nd_reshaped






import random, sys
import multiprocessing

from cmath import cos, exp, pi, sin, sqrt
from jax.scipy.linalg import expm
# from scipy.linalg import expm
from scipy.sparse import csr_matrix, csc_matrix
import scipy.sparse.linalg as ssla
import scipy

import numpy as np
from numpy import log
from numpy.linalg import matrix_power
np.set_printoptions(precision=6)
FLOATING_POINT_PRECISION = 1e-10

import jax
import jax.numpy as jnp
import jax.scipy.linalg as jsl

import matplotlib.pyplot as plt

# from quantum_simulation_recipe.bounds import *

def expH(H, t, use_jax=False):
    # # check H is Hermitian
    # if not np.allclose(H, H.conj().T):
    #     raise ValueError('H is not Hermitian')

    if use_jax: 
        if isinstance(H, np.ndarray):
            return jax.scipy.linalg.expm(-1j * t * H)
        else:
            return jax.scipy.linalg.expm(-1j * t * H.to_matrix())
    elif isinstance(H, csr_matrix):
        return scipy.sparse.linalg.expm(-1j * t * H)
    else:
        return scipy.linalg.expm(-1j * t * H)


def pf(h_list, t, r: int, order: int=2, use_jax=False, return_exact=False, verbose=False):
# def pf_r(h_list, t, r: int, order: int=2, use_jax=False, return_exact=False, verbose=False):
    if order == 1:
        list_U = [expH(herm, t/r, use_jax=use_jax) for herm in h_list]
        appro_U_dt = np.linalg.multi_dot(list_U)
        if isinstance(appro_U_dt, csr_matrix):
            appro_U = appro_U_dt**r
        else:
            appro_U = np.linalg.matrix_power(appro_U_dt, r)
    elif order == 2:
        list_U = [expH(herm, t/(2*r), use_jax=use_jax) for herm in h_list]
        if verbose: print('----expm Herm finished----')
        appro_U_dt_forward = np.linalg.multi_dot(list_U)
        appro_U_dt_reverse = np.linalg.multi_dot(list_U[::-1])
        # appro_U_dt = list_U[0] @ list_U[1]
        if verbose: print('----matrix product finished----')
        if isinstance(appro_U_dt_forward, csr_matrix):
            appro_U = (appro_U_dt_forward @ appro_U_dt_reverse)**r
        else:
            appro_U = np.linalg.matrix_power(appro_U_dt_reverse @ appro_U_dt_forward, r)
        if verbose: print('----matrix power finished----')
    else: 
        raise ValueError('higher order is not defined')

    if return_exact:
        exact_U = expH(sum(h_list), t)
        return appro_U, exact_U
    else:
        return appro_U

# def pf(list_herm, order, t):
#     # print('order: ', order)
#     if order == 1:
#         return unitary_matrix_product(list_herm, t)
#     elif order == 2:
#         forward_order_product = unitary_matrix_product(list_herm, t/2) 
#         reverse_order_product = unitary_matrix_product(list_herm[::-1], t/2)
#         return forward_order_product @ reverse_order_product
#         # return second_order_trotter(list_herm, t)
#     elif order > 0 and order!= 1 and order != 2 and order % 2 == 0:
#         p = 1 / (4 - 4**(1/(order-1)))
#         # print('p: ', p)
#         return matrix_power(pf(list_herm, order-2, p*t), 2) @ pf(list_herm, order-2, (1-4*p)*t) @ matrix_power(pf(list_herm, order-2, p*t), 2)
#     else:
#         raise ValueError('k is not defined')

# matrix product of a list of matrices
def unitary_matrix_product(list_herm_matrices, t=1):
    ''' 
    matrix product of a list of unitary matrices exp(itH)
    input: 
        list_herm_matrices: a list of Hermitian matrices
        t: time
    return: the product of the corresponding matrices
    '''
    product = expm(-1j * t * list_herm_matrices[0])
    for i in range(1, len(list_herm_matrices)):
        product = product @ expm(-1j * t * list_herm_matrices[i])

    return product

def matrix_product(list_U, t=1):
    # product = matrix_power(list_U[0], t)
    # for i in range(1, len(list_U)):
    #     product = matrix_power(list_U[i], t) @ product
    #     # product = product @ matrix_power(list_U[i], t)
    product = np.linalg.multi_dot([matrix_power(U, t) for U in list_U])
    return product

# def second_order_trotter(list_herm_matrices, t=1):
#     forward_order_product = unitary_matrix_product(list_herm_matrices, t/2) 
#     reverse_order_product = unitary_matrix_product(list_herm_matrices[::-1], t/2)

#     return forward_order_product @ reverse_order_product

def pf_U(list_U, order, t=1):
    # print('order: ', order)
    if order == 1:
        return matrix_product(list_U, t)
    elif order == 2:
        forward_order_product = matrix_product(list_U, t/2) 
        reverse_order_product = matrix_product(list_U[::-1], t/2)
        return forward_order_product @ reverse_order_product
    elif order > 0 and order != 1 and order != 2 and order % 2 == 0:
        p = 1 / (4 - 4**(1/(order-1)))
        # print('p: ', p)
        return matrix_power(pf_U(list_U, order-2, p*t), 2) @ pf_U(list_U, order-2, (1-4*p)*t) @ matrix_power(pf_U(list_U, order-2, p*t), 2)
    else:
        raise ValueError('k is not defined')


########### jax, mpi, sparse ###########
def jax_matrix_exponential(matrix):
    # return jsl.expm( matrix)
    return ssla.expm(matrix)
jax_matrix_exponential = jax.jit(jax.vmap(jax_matrix_exponential))

def sparse_multi_dot(sparse_matrices):
    '''
    计算一个列表中所有矩阵的乘积
    '''
    product = sparse_matrices[0]
    for matrix in sparse_matrices[1:]:
        product = product.dot(matrix)
    return product
    # return product.toarray()

vectorized_sparse_expm = jax.vmap(ssla.expm)

def mpi_sparse_expm(list_herms, t, r):
    pool = multiprocessing.Pool(processes=multiprocessing.cpu_count())
    list_unitaries = pool.map(ssla.expm, -1j * t / r * np.array(list_herms))
    # Close the pool of workers
    pool.close()
    pool.join()

    return list_unitaries

# from quantum_simulation_recipe.measure import op_error
# def sparse_trotter_error(list_herm: list, r: int, t: int) -> float:
#     print('-------sparse_trotter_error--------')
#     exact_U = ssla.expm(-1j * t * sum(list_herm))
#     # list_U = jax_matrix_exponential(jnp.array(-1j * t / (2*r) * np.array(list_herm)))
#     # list_U = vectorized_sparse_expm(-1j * t / (2*r) * np.array(list_herm))
#     # list_herm_scaled = np.array([-1j * t / (2*r) * herm for herm in list_herm])
#     # list_U = ssla.expm(list_herm_scaled) 
#     # list_U = [ssla.expm(-1j * t / (2*r) * herm) for herm in list_herm]
#     list_U = mpi_sparse_expm(list_herm, t, 2*r)
#     # list_U = jax_matrix_exponential(jnp.array([-1j * t / (2*r) * herm.toarray() for herm in np.array(list_herm)]))
#     list_U2 = [U**2 for U in list_U]
#     # trotter_error_list = op_error(exact_U, matrix_power(sparse_multi_dot(list_U2), r))
#     trotter_error_list = op_error(exact_U, sparse_multi_dot(list_U2)**r)
#     # trotter_error_list = op_error(exact_U, np.linalg.matrix_power(np.linalg.multi_dot(np.array(list_U2)), r))
#     # second-order trotter
#     trotter_error_list_2nd = op_error(exact_U, (sparse_multi_dot(list_U) @ sparse_multi_dot(list_U[::-1]))**r)
#     # trotter_error_list_2nd = op_error(exact_U, np.linalg.matrix_power(np.linalg.multi_dot(np.array(list_U)) @ np.linalg.multi_dot(np.array(list_U[::-1])), r))
    
#     return [trotter_error_list, trotter_error_list_2nd]



import numpy as np

import matplotlib as mpl
import matplotlib.pyplot as plt
import matplotlib.ticker as ticker

from matplotlib.colors import ListedColormap
import colorsys

SMALL_SIZE = 14
MEDIUM_SIZE = 18  #default 10
LARGE_SIZE = 24
# MARKER_SIZE = 10

plt.rc('font', size=MEDIUM_SIZE)  # controls default text sizes
plt.rc('axes', titlesize=LARGE_SIZE)  # fontsize of the axes title
plt.rc('axes', labelsize=LARGE_SIZE)  # fontsize of the x and y labels
plt.rc('xtick', labelsize=LARGE_SIZE)  # fontsize of the tick labels
plt.rc('ytick', labelsize=LARGE_SIZE)  # fontsize of the tick labels
plt.rc('legend', fontsize=MEDIUM_SIZE)  # legend fontsize
plt.rc('figure', titlesize=LARGE_SIZE)  # fontsize of the figure title

mpl.rcParams['font.family'] = 'sans-serif'  # 'Helvetica'
mpl.rcParams['axes.linewidth'] = 1.5
mpl.rcParams["xtick.direction"] = 'out' # 'out'
mpl.rcParams["ytick.direction"] = 'out'
mpl.rcParams['xtick.major.width'] = 1.5
mpl.rcParams['ytick.major.width'] = 1.5
mpl.rcParams['ytick.minor.width'] = 1.5
mpl.rcParams['lines.markersize'] = 11
mpl.rcParams['legend.frameon'] = True
mpl.rcParams['lines.linewidth'] = 1.5
# plt.rcParams['lines.markeredgecolor'] = 'k'
mpl.rcParams['lines.markeredgewidth'] = 1.5
mpl.rcParams['figure.dpi'] = 100
mpl.rcParams['figure.figsize'] = (8, 6)
mpl.rcParams['figure.autolayout'] = True
mpl.rcParams['axes.grid'] = True
mpl.rcParams['savefig.bbox'] = 'tight'
mpl.rcParams['savefig.transparent'] = True


# Function to lighten a color
def lighten_color(color, amount=0.3):
    # Convert color from hexadecimal to RGB
    if isinstance(color, str):
        r, g, b, a = tuple(int(color[i:i+2], 16) for i in (1, 3, 5, 7))
    else:
        r, g, b, a = color
        if 0 <= r <= 1 and 0 <= g <= 1 and 0 <= b <= 1 and 0 <= a <= 1:
            r, g, b, a = int(r*255), int(g*255), int(b*255), int(a*255)
        else: 
            raise ValueError('Color should be in hexadecimal or RGB format')
    # print(r, g, b, a)
    # Convert RGB to HLS
    h, l, s = colorsys.rgb_to_hls(r/255, g/255, b/255)
    # Lighten the luminance component
    l = min(1, l + amount)
    # Convert HLS back to RGB
    r, g, b = tuple(round(c * 255) for c in colorsys.hls_to_rgb(h, l, s))
    # Convert RGB back to hexadecimal
    new_color = f"#{r:02x}{g:02x}{b:02x}{a:02x}"
    return new_color

def set_color_cycle(color_cycle, alpha=0.3, mfc=False, edgecolor=False):
    color_cycle_light = [lighten_color(color, alpha) for color in color_cycle]
    if mfc:
        colors = mpl.cycler(mfc=color_cycle_light, color=color_cycle, markeredgecolor=color_cycle)
    else:
        if edgecolor:
            colors = mpl.cycler(color=color_cycle, markeredgecolor=color_cycle)
        else:
            colors = mpl.cycler(color=color_cycle)
    mpl.rc('axes', prop_cycle=colors)

default_color_cycle = ["#B65655FF", "#5471abFF", "#6aa66eFF", "#A66E6AFF"]
# set_color_cycle(default_color_cycle)
# mpl.rc('axes', grid=True, edgecolor='k', prop_cycle=colors)
# mpl.rcParams['axes.prop_cycle'] = colors
# mpl.rcParams['lines.markeredgecolor'] = 'C'

from colorspace import sequential_hcl
class GradColors:
    def __init__(self, rate_num):
        self.rate_num = rate_num
        # https://colorspace.r-forge.r-project.org/reference/hcl_palettes.html
        self.purple = mpl.colors.ListedColormap(sequential_hcl("Purples")(rate_num+1)[:-1][::-1], name='from_list', N=None) 
        self.red = mpl.colors.ListedColormap(sequential_hcl("Reds")(rate_num+1)[:-1][::-1], name='from_list', N=None) 
        self.green = mpl.colors.ListedColormap(sequential_hcl("Greens")(rate_num+1)[:-1][::-1], name='from_list', N=None) 
        self.blue = mpl.colors.ListedColormap(sequential_hcl("Blues")(rate_num+1)[:-1][::-1], name='from_list', N=None) 
        self.orange = mpl.colors.ListedColormap(sequential_hcl("Oranges")(rate_num+1)[:-1][::-1], name='from_list', N=None) 
        self.mint = mpl.colors.ListedColormap(sequential_hcl("Mint")(rate_num+1)[:-1][::-1], name='from_list', N=None)
        # self.return_colors()

    def get_colors(self, c: str):
        return mpl.colors.ListedColormap(sequential_hcl(c)(self.rate_num+1)[:-1][::-1], name='from_list', N=None) 


from scipy.optimize import curve_fit
from math import ceil, floor, log, exp

def linear_loglog_fit(x, y, verbose=False):
    # Define the linear function
    def linear_func(x, a, b):
        return a * x + b

    log_x = np.array([log(n) for n in x])
    log_y = np.array([log(cost) for cost in y])
    # Fit the linear function to the data
    params, covariance = curve_fit(linear_func, log_x, log_y)
    # Extract the parameters
    a, b = params
    # Predict y values
    y_pred = linear_func(log_x, a, b)
    # Print the parameters
    if verbose: print('Slope (a):', a, '; Intercept (b):', b)
    exp_y_pred = [exp(cost) for cost in y_pred]

    return exp_y_pred, a, b

def plot_fit(ax, x, y, var='t', x_offset=1.07, y_offset=1.0, label='', ext_x=[], linestyle='k--', linewidth=1.5, fontsize=MEDIUM_SIZE, verbose=True):
    y_pred_em, a_em, b_em = linear_loglog_fit(x, y)
    if verbose: print(f'a_em: {a_em}; b_em: {b_em}')
    if abs(a_em) < 1e-3: 
        text_a_em = "{:.2f}".format(round(abs(a_em), 4))
    else:
        text_a_em = "{:.2f}".format(round(a_em, 4))

    if ext_x != []: x = ext_x
    y_pred_em = [exp(cost) for cost in a_em*np.array([log(n) for n in x]) + b_em]
    if label =='':
        ax.plot(x, y_pred_em, linestyle, linewidth=linewidth)
    else:
        ax.plot(x, y_pred_em, linestyle, linewidth=linewidth, label=label)
    ax.annotate(r'$O(%s^{%s})$' % (var, text_a_em), xy=(x[-1], np.real(y_pred_em)[-1]), xytext=(x[-1]*x_offset, np.real(y_pred_em)[-1]*y_offset), fontsize=fontsize)

    return a_em, b_em

def ax_set_text(ax, x_label, y_label, title=None, legend='best', xticks=None, yticks=None, grid=None, log='', ylim=None):
    ax.set_xlabel(x_label)
    ax.set_ylabel(y_label)
    if title: ax.set_title(title)
    if legend: ax.legend(loc=legend)

    if log == 'x': 
        ax.set_xscale('log')
    elif log == 'y':
        ax.set_yscale('log')
    elif log == 'xy':
        # ax.set_xscale('log')
        # ax.set_yscale('log')
        ax.loglog()
    else:
        pass

    if grid: ax.grid()  
    if ylim: ax.set_ylim([ylim[0]*0.85, ylim[1]*1.15])

    if xticks is not None: 
        ax.set_xticks(xticks)
        ax.get_xaxis().set_major_formatter(mpl.ticker.ScalarFormatter())

    if yticks is not None: 
        ax.set_yticks(yticks)
        ax.get_yaxis().set_major_formatter(mpl.ticker.ScalarFormatter())

# def data_plot(x, y, marker, label, alpha=1, linewidth=1, loglog=True, markeredgecolor='black'):
#     if loglog:
#         plt.loglog(x, y, marker, label=label, linewidth=linewidth, markeredgecolor=markeredgecolor, markeredgewidth=0.5, alpha=alpha)
#     else:
#         plt.plot(x, y, marker, label=label, linewidth=linewidth, markeredgecolor=markeredgecolor, markeredgewidth=0.5, alpha=alpha)

def plot_evo(ax, t_list, y_list, marker, color='', title='', xlabel='', ylabel='', label='', markersize=10, markeredgewidth=1.5, inset=False):
    if color == '':
        ax.plot(t_list, y_list, marker, label=label, markersize=markersize, markeredgewidth=markeredgewidth)
        # ax.plot(t_list, y_list, '-', markersize=5)
        # ax.plot(t_list, y_list, 'o', label=label, markersize=5)
        # ax.plot(t_list, y_list, marker, label=label, markeredgecolor='k', markeredgewidth=0.4, markersize=5)
    else:
        ax.plot(t_list, y_list, marker, color=color, label=label, markeredgecolor=color, markeredgewidth=markeredgewidth, markersize=markersize, mfc=lighten_color(color, 0.3))
        # ax.plot(t_list, y_list, marker, color=color, label=label, markeredgecolor=color, markeredgewidth=0.4, markersize=markersize, mfc=color[:-2]+"80")
    if not inset: 
        ax.set_title(title)
        ax.set_xlabel(xlabel)
        ax.set_ylabel(ylabel)
    if title  != '': ax.set_title(title)
    if xlabel != '': 
        ax.set_xlabel(xlabel)
    # else:
    #     ax.set_xticks([]) 
    if ylabel != '': ax.set_ylabel(ylabel)
    # else:
    #     ax.set_xticks([])

def letter_annotation(axes, x_offset, y_offset, letters, fontsize=14):
    # https://towardsdatascience.com/a-guide-to-matplotlib-subfigures-for-creating-complex-multi-panel-figures-70fa8f6c38a4
    for letter in letters:
        axes[letter].text(x_offset, y_offset, f'({letter})', transform=axes[letter].transAxes, size=fontsize, weight='bold')


def matrix_plot(M):
    fig, ax = plt.subplots()
    real_matrix = np.real(M)
    # Plot the real part using a colormap
    ax.imshow(real_matrix, cmap='RdYlBu', interpolation='nearest', origin='upper')
    # Create grid lines
    ax.grid(True, which='both', color='black', linewidth=1)
    # Add color bar for reference
    cbar = plt.colorbar(ax.imshow(real_matrix, cmap='RdYlBu', interpolation='nearest', origin='upper'), ax=ax, orientation='vertical')
    cbar.set_label('Real Part')
    # Add labels to the x and y axes
    plt.xlabel('X-axis')
    plt.ylabel('Y-axis')
    # Show the plot
    plt.title('Complex Matrix with Grid')
    plt.show()


from cmath import exp
from scipy import sparse
from scipy.linalg import expm
from numpy.linalg import matrix_power
import scipy.sparse.linalg as ssla

import numpy as np
from quantum_simulation_recipe.measure import commutator, norm

# def commutator(A, B):
#     return A @ B - B @ A

# def norm(A, ord='spectral'):
#     if ord == 'fro':
#         return np.linalg.norm(A)
#     elif ord == 'spectral':
#         return np.linalg.norm(A, ord=2)
#     elif ord == '4':
#         return np.trace(A @ A.conj().T @ A @ A.conj().T)**(1/4)
#     else:
#         # raise ValueError('norm is not defined')
#         return np.linalg.norm(A, ord=ord)

def analytic_bound(H, k, t, r):
    L = len(H)
    Lambda = max([norm(h) for h in H])

    return (2 * L * 5**(k-1) * Lambda * t)**(2*k+1)/(3*r**(2*k)) * exp((2*L*5**(k-1)*Lambda*t)/r)

def interference_bound(H, t, r):
    # Layden_2022_First-Order Trotter Error from a Second-Order Perspective
    try:
        assert len(H) == 2
    except:
        raise ValueError('The Hamiltonian contains not exactly 2 terms')

    h1 = H[0]
    h2 = H[1]
    C1 = min(norm(h1), norm(h2))
    C2 = 0.5 * norm(commutator(h1, h2))
    S = [norm(commutator(h1, commutator(h1, h2))), norm(commutator(h2, commutator(h2, h1)))]
    C3 = 1 / 12 * (min(S) + 0.5 * max(S))
    e1 = C1 * t / r
    e2 = C2 * t**2 / r
    e3 = C3 * t**3 / r**2
    bound = min(e2, e1 + e3, 2)
    # bound = min(e2, e1 + e3, 2 * len(h1))

    return bound, e1, e2, e3

def triangle_bound(h, k, t, r):
    L = len(h)
    if k == 1:
        if L == 2:
            raise ValueError('k=1 is not defined for L=2')
        elif L == 3:
            c = norm(commutator(h[0], h[1])) + norm(commutator(h[1], h[2])) + norm(commutator(h[2], h[0]))
            error = c * t**2 / (2*r) 
    return error

def tight_bound(h, order: int, t: float, r: int, type='spectral', verbose=False):
    L = len(h)
    d = h[0].shape[0]
    if order == 1:
        a_comm = 0
        for i in range(0, L):
            temp = np.zeros(h[0].shape, dtype=complex)
            for j in range(i + 1, L):
                temp += commutator(h[i], h[j])
            a_comm += norm(temp, ord=type)

        if type == 'spectral':
            error = a_comm * t**2 / (2*r)
        elif type == 'fro':
            error = a_comm * t**2 / (2*r*np.sqrt(d))
        else:
            raise ValueError(f'type={type} is not defined')
    elif order == 2:
        c1 = 0
        c2 = 0
        for i in range(0, L):
            temp = np.zeros(h[0].shape, dtype=complex)
            for j in range(i + 1, L):
                temp += h[j]
            # h_sum3 = sum(h[k] for k in range(i+1, L))
            # print(h_sum3.shape)
            # h_sum2 = sum(h[k] for k in range(i+1, L))
            c1 += norm(commutator(temp, commutator(temp, h[i])), ord=type) 
            # c1 = norm(commutator(h[0]+h[1], commutator(h[1]+h[2], h[0]))) + norm(commutator(h[2], commutator(h[2], h[1])))
            # c2 = norm(commutator(h[0], commutator(h[0],h[1]+h[2]))) + norm(commutator(h[1], commutator(h[1], h[2])))
            c2 += norm(commutator(h[i], commutator(h[i], temp)), ord=type)
        if type == 'spectral':
            error = c1 * t**3 / r**2 / 12 + c2 *  t**3 / r**2 / 24 
        elif type == 'fro':
            # print(c1, c2)
            error = c1 * t**3 / r**2 / 12 / np.sqrt(d) + c2 *  t**3 / r**2 / 24 / np.sqrt(d)
            # print('random input:', error)
        elif type == '4':
            error = c1 * t**3 / r**2 / 12 / d**(1/4) + c2 *  t**3 / r**2 / 24 / d**(1/4)
        else:
            raise ValueError(f'type={type} is not defined')
    else: 
        raise ValueError(f'higer order (order={order}) is not defined')

    return error

def analytic_loose_commutator_bound_parity(n, J, h, dt, pbc=False, verbose=False):
    if pbc:
        c1 = 16*J**2*h*(n) + 4*J**2*h*(n)
        c2 = 8*(n)*J**2*h
    else:
        # c1 = 16*J**2*h*(n-1) + 4*J**2*h*(n-2)
        # c2 = 8*(n-1)*J**2*h
        if n % 2 == 1:
            c1 = 4*J*h**2*(n-1) + 4*J**2*h*(n-1)
            c2 = 4*J*h**2*(n-1) + 4*J**2*h*(n-2)
        else:
            c1 = 4*J*h**2*(n-1) + 4*J**2*h*(n-2)
            c2 = 4*J*h**2*(n-1) + 4*J**2*h*(n-1)

    if verbose: print(f'c1 (analy)={c1}, c2={c2}')
    analytic_error_bound = c1 * dt**3 / 12 + c2 * dt**3 / 24
    return analytic_error_bound, c1, c2

def analytic_loose_commutator_bound_xyz(n, J, h, dt, pbc=False, verbose=False):
    if pbc:
        c1 = 16*J**2*h*(n) + 4*J**2*h*(n)
        c2 = 8*(n)*J**2*h
    else:
        c1 = 8*J*h**2*2*(n-1) 
        c2 =  8*J**2*h*(2*(n-1)-2) + 4*J**2*h*(2)

    if verbose: print(f'c1 (analy)={c1}, c2={c2}')
    analytic_error_bound = c1 * dt**3 / 12 + c2 * dt**3 / 24
    return analytic_error_bound, c1, c2

def analy_st_loose_bound(r, n, J, h, t, ob_type='single', group='parity'):
    if group == 'parity':
        return 2 * analytic_loose_commutator_bound_parity(n, J, h, t/r)[0] * r
    elif group == 'xyz':
        return 2 * analytic_loose_commutator_bound_xyz(n, J, h, t/r)[0] * r
    else:
        raise ValueError(f'group={group} not recognized')

def analy_st_bound(r, n, J, h, t, ob_type='single'):
    if ob_type == 'single':
        return 2 * analytic_loose_commutator_bound(n, J, h, t/r) * r
    elif ob_type == 'multi':
        # return 2 * analytic_loose_commutator_bound(n, J, h, t/r) * r  * n
        return 2 * analytic_loose_commutator_bound(n, J, h, t/r) * r 
    else:
        raise ValueError('ob_type should be either single or multi')

def analy_lc_bound(r, n, J, h, t, ob_type='single', verbose=False):
    err_bound = 0
    # for i in range(1, r+1):
    for i in range(1, r+2):
        if ob_type == 'single':
            n_lc = min(i*2, n)
            err_bound += 2 * analytic_loose_commutator_bound(n_lc, J, h, t/r, verbose) 
        elif ob_type == 'multi':   
            for j in range(0, n):
                n_lc = min(min(n-j, i*2) + min(j, 2*i), n)
                # err_bound += 2 * analytic_loose_commutator_bound(n_lc, J, h, t/r, verbose) 
                err_bound += 2 * analytic_loose_commutator_bound(n_lc, J, h, t/r, verbose) / n
        else:
            raise ValueError('ob_type should be either single or multi')

    return err_bound

def analytic_loose_commutator_bound(n, J, h, dt, pbc=False, verbose=False):
    if pbc:
        c1 = 16*J**2*h*(n) + 4*J**2*h*(n)
        c2 = 8*(n)*J**2*h
    else:
        # c1 = 16*J**2*h*(n-1) + 4*J**2*h*(n-2)
        # c2 = 8*(n-1)*J**2*h
        if n % 2 == 1:
            c1 = 4*J*h**2*(n-1) + 4*J**2*h*(n-1)
            c2 = 4*J*h**2*(n-1) + 4*J**2*h*(n-2)
        else:
            c1 = 4*J*h**2*(n-1) + 4*J**2*h*(n-2)
            c2 = 4*J*h**2*(n-1) + 4*J**2*h*(n-1)


    if verbose: print(f'c1 (analy)={c1}, c2={c2}')
    analytic_error_bound = c1 * dt**3 / 12 + c2 * dt**3 / 24
    return analytic_error_bound

from quantum_simulation_recipe.spin import *

# def relaxed_commutator_bound(n, J, h, dt, verbose=False):
#     hnn = Nearest_Neighbour_1d(n, Jx=J, Jy=J, Jz=J, hx=h, hy=0, hz=0, pbc=False, verbose=False)
#     h_list = hnn.ham_par
def relaxed_commutator_bound(n, cmm_data, dt, verbose=False):
    if verbose: print(f'c1 (relax)={cmm_data["c1"][n-1]}, c2={cmm_data["c2"][n-1]}')
    relaxed_error_bound = cmm_data['c1'][n-1] * dt**3 / 12 + cmm_data['c2'][n-1] * dt**3 / 24
    return relaxed_error_bound

# def relaxed_st_bound(r, n, cmm_data, t, ob_type='singl'):
def relaxed_st_bound(r, n, h, t, h_group=[], ob_type='singl'):
    if h_group == []:
        h_list = h.ham_par
        # h_list = h.ham_xyz[:2]
    else:
        h_list = h_group
    # h_list = lc_group(h, 0, 0, 2*n+2)[:2]
    dt = t/r
    if ob_type == 'singl':
        c1_cmm = commutator(h_list[1], commutator(h_list[1], h_list[0]).simplify()).simplify()
        c2_cmm = commutator(h_list[0], commutator(h_list[0], h_list[1]).simplify()).simplify()
        c1 = np.linalg.norm(c1_cmm.coeffs, ord=1)
        c2 = np.linalg.norm(c2_cmm.coeffs, ord=1)
        if c1 >= c2:
            return 2 * (c1 * dt**3 / 12 + c2 * dt**3 / 24) * r
        else:
            return 2 * (c2 * dt**3 / 12 + c1 * dt**3 / 24) * r
    #     return 2 * relaxed_commutator_bound(n, cmm_data, t/r) * r
    # elif ob_type == 'multi':
    #     # return 2 * analytic_loose_commutator_bound(n, J, h, t/r) * r  * n
    #     return 2 * relaxed_commutator_bound(n, cmm_data, t/r) * r 
    else:
        raise ValueError('ob_type should be either single or multi')

def relaxed_lc_bound(r, n, cmm_data, t, ob_type='singl', verbose=False):
    err_bound = 0
    for i in range(1, r+2):
    # for i in range(1, r+1):
        if ob_type == 'singl':
            n_lc = min(i*2, n)
            err_bound += 2 * relaxed_commutator_bound(n_lc, cmm_data, t/r, verbose=verbose) 
        elif ob_type == 'multi':   
            for j in range(0, n):
                n_lc = min(min(n-j, i*2) + min(j, 2*i), n)
                # err_bound += 2 * analytic_loose_commutator_bound(n_lc, J, h, t/r, verbose) 
                err_bound += 2 * relaxed_commutator_bound(n_lc, cmm_data, t/r, verbose=verbose) / n
        else:
            raise ValueError('ob_type should be either single or multi')

    return err_bound

# # def lc_tail_bound(r, n, tail_cmm_data, t, ob_type='singl', verbose=False):
# def lc_tail_bound(r, n, tail_cmm_data, t, ob_type='singl', verbose=True):
#     err_bound = 0
#     dt = t/r
#     if verbose: print(f'index={n}, r={r}')
#     if verbose: print(tail_cmm_data['n'][n-1])
#     if ob_type == 'singl':
#         err_bound = 2 * (tail_cmm_data['c1_z'][n-1][r-1][0] * dt**3 / 12 + tail_cmm_data['c2_z'][n-1][r-1][0] * dt**3 / 24)
#     elif ob_type == 'multi_z':
#         err_bound = 2 * (sum(tail_cmm_data['c1_z'][n-1][r-1]) * dt**3 / 12 + sum(tail_cmm_data['c2_z'][n-1][r-1]) * dt**3 / 24 ) / tail_cmm_data['n'][n-1]
#     elif ob_type == 'multi_zz':
#         err_bound = 2 * (sum(tail_cmm_data['c1_zz'][n-1][r-1]) * dt**3 / 12 + sum(tail_cmm_data['c2_zz'][n-1][r-1]) * dt**3 / 24) / (tail_cmm_data['n'][n-1]-1)
#     else:
#         raise ValueError('ob_type should be either single or multi')

#     return err_bound

def lc_group(h, right, left, step, verbose=False, legacy=False):
    if verbose: print(f'n={h.n}, right={right}, left={left}, step={step}')
    tail_tuples = []
    right_range = list(range(0, right-step))
    left_range = list(range(left+step, h.n-1))
    if verbose: print(right_range, left_range)
    all_tuples = [h.x_tuples, h.y_tuples, h.z_tuples, h.xx_tuples, h.yy_tuples, h.zz_tuples]
    if verbose: print(all_tuples)
    for i in right_range:
        for tuple in all_tuples:
            tail_tuples.append(tuple[i])
            # if verbose: print(tuple[i])

    for i in left_range:
        # for tuple in all_tuples:
        tail_tuples.append(h.xx_tuples[i])
        tail_tuples.append(h.yy_tuples[i])
        tail_tuples.append(h.zz_tuples[i])
        tail_tuples.append(h.x_tuples[i+1])
        tail_tuples.append(h.y_tuples[i+1])
        tail_tuples.append(h.z_tuples[i+1])
            # tail_tuples.append(tuple[i])
            # if verbose: print(tuple[i])

    # if left_range != []:
    #     tail_tuples.append(h.x_tuples[-1])
    #     tail_tuples.append(h.y_tuples[-1])
    #     tail_tuples.append(h.z_tuples[-1])
    # print([*h.x_tuples, *h.y_tuples, *h.z_tuples, *h.xx_tuples, *h.yy_tuples, *h.zz_tuples]-tail_terms)
    odd_lc_tuples, even_lc_tuples = [], []
    for i in list(range(max(right-step, 0), right))[::-1][::2]:
        odd_lc_tuples.append(h.xx_tuples[i])
        odd_lc_tuples.append(h.yy_tuples[i])
        odd_lc_tuples.append(h.zz_tuples[i])
        odd_lc_tuples.append(h.x_tuples[i])
        odd_lc_tuples.append(h.y_tuples[i])
        odd_lc_tuples.append(h.z_tuples[i])

    if verbose: print(list(range(left, min(left+step, h.n-1)))[::2])
    for i in list(range(left, min(left+step, h.n-1)))[::2]:
        odd_lc_tuples.append(h.xx_tuples[i])
        odd_lc_tuples.append(h.yy_tuples[i])
        odd_lc_tuples.append(h.zz_tuples[i])
        odd_lc_tuples.append(h.x_tuples[i+1])
        odd_lc_tuples.append(h.y_tuples[i+1])
        odd_lc_tuples.append(h.z_tuples[i+1])

    for i in range(right, left):
        even_lc_tuples.append(h.xx_tuples[i])
        even_lc_tuples.append(h.yy_tuples[i])
        even_lc_tuples.append(h.zz_tuples[i])
        even_lc_tuples.append(h.x_tuples[i])
        even_lc_tuples.append(h.y_tuples[i])
        even_lc_tuples.append(h.z_tuples[i])
    even_lc_tuples.append(h.x_tuples[left])
    even_lc_tuples.append(h.y_tuples[left])
    even_lc_tuples.append(h.z_tuples[left])

    for i in list(range(max(right-step, 0), right-1))[::-1][::2]:
        even_lc_tuples.append(h.xx_tuples[i])
        even_lc_tuples.append(h.yy_tuples[i])
        even_lc_tuples.append(h.zz_tuples[i])
        even_lc_tuples.append(h.x_tuples[i])
        even_lc_tuples.append(h.y_tuples[i])
        even_lc_tuples.append(h.z_tuples[i])

    for i in list(range(left+1, min(left+step, h.n-1)))[::2]:
        even_lc_tuples.append(h.xx_tuples[i])
        even_lc_tuples.append(h.yy_tuples[i])
        even_lc_tuples.append(h.zz_tuples[i])
        even_lc_tuples.append(h.x_tuples[i+1])
        even_lc_tuples.append(h.y_tuples[i+1])
        even_lc_tuples.append(h.z_tuples[i+1])
    # for i in
    all_pauli = SparsePauliOp.from_sparse_list([*h.xx_tuples, *h.yy_tuples, *h.zz_tuples, *h.x_tuples, *h.y_tuples, *h.z_tuples], h.n).simplify() 
    lc_pauli = SparsePauliOp.from_sparse_list([*odd_lc_tuples, *even_lc_tuples, *tail_tuples], h.n).simplify()
    difference = (all_pauli - lc_pauli).simplify()

    if len(difference.coeffs) == 1 and difference.coeffs[0] == 0:
        if verbose: print('Success!')
    else:
        print('Lightcone decomposition error!: ', difference)
        # print(set([*odd_lc_tuples, *even_lc_tuples, *tail_tuples])-set([*h.x_tuples, *h.y_tuples, *h.z_tuples, *h.xx_tuples, *h.yy_tuples, *h.zz_tuples]))
        # print('', [*odd_lc_tuples, *even_lc_tuples, *tail_tuples])
        # print('', [*h.x_tuples, *h.y_tuples, *h.z_tuples, *h.xx_tuples, *h.yy_tuples, *h.zz_tuples])
        # raise Exception( e.args )

    even_lc_terms = SparsePauliOp.from_sparse_list(even_lc_tuples, h.n).simplify()
    odd_lc_terms  = SparsePauliOp.from_sparse_list(odd_lc_tuples, h.n).simplify()
    tail_lc_terms = SparsePauliOp.from_sparse_list(tail_tuples, h.n).simplify()
    # print(h.even_terms)
    # print(h.x_tuples)
    # return [even_lc_tuples, odd_lc_tuples, tail_tuples]
    
    new_even_lc_tuples, new_odd_lc_tuples = [], []
    for item in even_lc_tuples + odd_lc_tuples:
        if verbose: print(item)
        if item[1][0] % 2 == 0 and item[2] != 0:
            if verbose: print('even: ', item)
            new_even_lc_tuples.append(item)
        elif item[1][0] % 2 == 1 and item[2] != 0:
            if verbose: print('odd: ', item)
            new_odd_lc_tuples.append(item)

    temp = []
    if len(new_even_lc_tuples) % 2 == 1:
        for item in new_even_lc_tuples:
            if len(item[0]) == 1:
                temp.append(item)
        if verbose: print('single even: ', sorted(temp, key=lambda x: x[1][0]))
        boundary_element = sorted(temp, key=lambda x: x[1][0])[-1]
        new_even_lc_tuples.remove(boundary_element)
        new_odd_lc_tuples.append(boundary_element)
    else:
        for item in new_odd_lc_tuples:
            if len(item[0]) == 1:
                temp.append(item)
        if verbose: print('single odd: ', sorted(temp, key=lambda x: x[1][0]))
        boundary_element = sorted(temp, key=lambda x: x[1][0])[-1]
        new_odd_lc_tuples.remove(boundary_element)
        new_even_lc_tuples.append(boundary_element)

    if verbose:
        print("even: ", SparsePauliOp.from_sparse_list(new_even_lc_tuples, h.n))
        print("odd:  ", SparsePauliOp.from_sparse_list(new_odd_lc_tuples, h.n))
        print("tail: ", SparsePauliOp.from_sparse_list(tail_tuples, h.n))

    all_pauli = SparsePauliOp.from_sparse_list([*new_odd_lc_tuples, *new_even_lc_tuples, *tail_tuples], h.n).simplify() 
    lc_pauli = SparsePauliOp.from_sparse_list([*odd_lc_tuples, *even_lc_tuples, *tail_tuples], h.n).simplify()
    difference = (all_pauli - lc_pauli).simplify()
    if len(difference.coeffs) == 1 and difference.coeffs[0] == 0:
        # print('Success!')
        if verbose: print('Success!')
    else:
        print('Lightcone decomposition error!: ', difference)

    if legacy:
        return [even_lc_terms, odd_lc_terms, tail_lc_terms]
    else:
        even_lc_terms = SparsePauliOp.from_sparse_list(new_even_lc_tuples, h.n).simplify()
        odd_lc_terms  = SparsePauliOp.from_sparse_list(new_odd_lc_tuples, h.n).simplify()

        return [even_lc_terms, odd_lc_terms, tail_lc_terms]


def nested_commutator_norm(h_list):
    if len(h_list) == 2:
        c1_cmm = commutator(h_list[1], commutator(h_list[1], h_list[0]).simplify()).simplify()
        c2_cmm = commutator(h_list[0], commutator(h_list[0], h_list[1]).simplify()).simplify()
        c1_cmm_norm = np.linalg.norm(c1_cmm.coeffs, ord=1)
        c2_cmm_norm = np.linalg.norm(c2_cmm.coeffs, ord=1)
    elif len(h_list) == 3:
        c1_cmm_0 = commutator(h_list[1]+h_list[2], commutator(h_list[1]+h_list[2], h_list[0]).simplify()).simplify() 
        c1_cmm_1 = commutator(h_list[2], commutator(h_list[2], h_list[1]).simplify()).simplify()
        c2_cmm_0 = commutator(h_list[0], commutator(h_list[0], h_list[1]+h_list[2]).simplify()).simplify() 
        c2_cmm_1 = commutator(h_list[1], commutator(h_list[1], h_list[2]).simplify()).simplify()
        c1_cmm_norm = np.linalg.norm(c1_cmm_0.coeffs, ord=1) + np.linalg.norm(c1_cmm_1.coeffs, ord=1)
        c2_cmm_norm = np.linalg.norm(c2_cmm_0.coeffs, ord=1) + np.linalg.norm(c2_cmm_1.coeffs, ord=1)
    else:
        raise Exception('Invalid number of terms in the Hamiltonian list')

    return c1_cmm_norm, c2_cmm_norm

import jax
import jax.numpy as jnp
from jax import jit, vmap

import multiprocess as mp 
# import multiprocessing as mp

def lc_tail_bound(r, n, h, t, ob_type='singl', right=0, left=0, verbose=True):
    err_bound = 0
    dt = t/r
    if verbose: print(f'index={n}, r={r}')

    PROCESSES = 9
    # j_values = jnp.arange(1, r+1)
    if ob_type == 'singl':
        # @jit
        def process_item(j):
            h_list_z = lc_group(h, right, left, 2*j, verbose=False, legacy=True)
            c1_cmm_z, c2_cmm_z = nested_commutator_norm(h_list_z)
            return 2 * (c1_cmm_z * dt**3 / 12 + c2_cmm_z * dt**3 / 24)

        with mp.Pool(PROCESSES) as pool:
            results = pool.map(process_item, range(1, r+1))

        err_bound = sum(results)

        # process_item_vmap = vmap(process_item)
        # results = jax.device_get(process_item_vmap(j_values))
        # err_bound = jnp.sum(results)
        # err_bound = 0
        # for j in range(1, r+1):
        #     h_list_z  = lc_group(h, right, left, 2*j, verbose=False, legacy=True)
        #     c1_cmm_z, c2_cmm_z = nested_commutator_norm(h_list_z) 
        #     err_bound += 2 * (c1_cmm_z * dt**3 / 12 + c2_cmm_z * dt**3 / 24)
        # # err_bound = 2 * r * (c1_cmm_z * dt**3 / 12 + c2_cmm_z * dt**3 / 24)
    elif ob_type == 'multi_z':
        # err_bound = 0
        # for j in range(1, r+1):
        #     h_list_z_list = [lc_group(h, i, i, 2*j+2, verbose=False) for i in range(0, n)]
        #     # h_list_z_list = [lc_group(h, i, i, 4*r+1, False) for i in range(0, n)]
        #     # h_list_z_list = [lc_group(h, i, i, 2*r, False) for i in range(0, n)]
        #     tail_cmm_data = np.array([nested_commutator_norm(h_list_z) for h_list_z in h_list_z_list])
        #     err_bound += 2 * (sum(tail_cmm_data[:, 0])*dt**3/12 + sum(tail_cmm_data[:, 1])*dt**3/24) / n
        def process_item(j):
            h_list_z_list = [lc_group(h, i, i, 2*j+2, verbose=False) for i in range(0, n)]
            tail_cmm_data = np.array([nested_commutator_norm(h_list_z) for h_list_z in h_list_z_list])
            return 2 * (sum(tail_cmm_data[:, 0])*dt**3/12 + sum(tail_cmm_data[:, 1])*dt**3/24) / n

        with mp.Pool(PROCESSES) as pool:
            results = pool.map(process_item, range(1, r+1))

        err_bound = sum(results)
    elif ob_type == 'multi_zz':
        def process_item(j):
            h_list_zz_list = [lc_group(h, i, i+1, 2*j+2, verbose=False) for i in range(0, n-1)]
            tail_cmm_data = np.array([nested_commutator_norm(h_list_zz) for h_list_zz in h_list_zz_list])
            return 2 * (sum(tail_cmm_data[:, 0])*dt**3/12 + sum(tail_cmm_data[:, 1])*dt**3/24) / (n-1)

        with mp.Pool(PROCESSES) as pool:
            results = pool.map(process_item, range(1, r+1))

        err_bound = sum(results)
        # err_bound = 0
        # for j in range(1, r+1):
        #     h_list_zz_list = [lc_group(h, i, i+1, 2*j+2, verbose=False) for i in range(0, n-1)]
        #     # h_list_zz_list = [lc_group(h, i, i+1, 2*r, False) for i in range(0, n-1)]
        #     tail_cmm_data = np.array([nested_commutator_norm(h_list_zz) for h_list_zz in h_list_zz_list])
        #     err_bound += 2 * (sum(tail_cmm_data[:, 0])*dt**3/12 + sum(tail_cmm_data[:, 1])*dt**3/24) / (n-1)
        # # err_bound = 2 * (sum(tail_cmm_data['c1_zz'][n-1][r-1]) * dt**3 / 12 + sum(tail_cmm_data['c2_zz'][n-1][r-1]) * dt**3 / 24) / (n-1)
    else:
        raise ValueError('ob_type should be either single or multi')

    return err_bound


# def lc_nested_commutator_norm(i, h, r):
#     h_list_z  = lc_group(h, i, i, 2*r, False)
#     c1_cmm_z, c2_cmm_z = nested_commutator_norm(h_list_z) 

#     h_list_zz = lc_group(h, i, i+1, 2*r, False)
#     c1_cmm_zz, c2_cmm_zz = nested_commutator_norm(h_list_zz) 

#     return c1_cmm_z, c2_cmm_z, c1_cmm_zz, c2_cmm_zz

# def process_data(i):
#     # print(np.log(exp(i)))
#     # random matrix of dim i
#     m = np.random.rand(i, i)
#     m_norm = np.linalg.norm(m, ord=1)
#     print(i, m_norm)
#     return m_norm

def exp_count(r, n_qubits, factor, method, k=1):
    n_terms = factor * n_qubits
    if method == 'LC':
        exp_count = 0
        for i in range(1, r+1):
            # print('i: ', i)
            if i <= int((n_qubits-k)/2):
                exp_count += (3*k + 6 * i) * factor / 2
                # exp_count += (2*factor * i - 1+k) * 2    
                # exp_count += (4 * i - 1) * 2    
            else:
                exp_count += n_terms * 1.5
    elif method == 'ST':
        exp_count = 1.5 * n_terms * r

    return exp_count


def four_norm(ob):
    ob = ob.to_matrix()
    dim = ob.shape[0]
    # print('d', d)
    return (np.trace(ob @ ob @ ob @ ob)/dim)**(1/4)




def binary_search_r(r_start, r_end, epsilon, error_measure, step=1, comment='', verbose=False):
    # print(f'----binary search r ({error_measure.__name__})----')
    print(f'----[{comment}] binary search r (r_start={r_start}, r_end={r_end})----')
    while error_measure(r_end) > epsilon:
        print("the initial r_end is too small, increase it by 10 times.")
        r_end *= 10

    if error_measure(r_start) <= epsilon:
        r = r_start
    else: 
        while r_start < r_end - step: 
            r_mid = int((r_start + r_end) / 2)
            if error_measure(r_mid) > epsilon:
                r_start = r_mid
            else:
                r_end = r_mid
            if verbose: print('r_start:', r_start, '; r_end:', r_end)
        r = r_end
    if verbose: print('r:', r, '; err: ', error_measure(r))
    return r


# def search_r_for_error(r_start, r_end, epsilon, t, list_herm, k, norm='spectral', verbose=False):
#     tol = r_end - r_start
#     exact_U = expm(-1j * t * sum(list_herm))
#     # binary search from r_start to r_end
#     while tol > 2:
#         r = int((r_start + r_end) / 2)
#         err = op_error(matrix_power(pf(list_herm, k, t=t/r), r), exact_U, norm)
#         # if k == 1:
#         #     err = op_error(matrix_power(unitary_matrix_product(list_herm, t=t/r), r), exact_U, norm)
#         # elif k == 2:
#         #     err = op_error(matrix_power(second_order_trotter(list_herm, t=t/r), r), exact_U, norm)
#         # elif k != 2 and k > 1 and k % 2 == 0:
#         #     err = op_error(matrix_power(high_order_trotter(list_herm, k, t=t/r), r), exact_U, norm)
#         # else:
#         #     raise ValueError('k is not defined')

#         if err > epsilon:
#             r_start = r
#         else:
#             r_end = r
#         tol = abs(r_end - r_start)
#     if verbose: print('err: ', err)
#     return r


def normalize(data):
    s = sum(a**2 for a in data)
    return [a**2/s for a in data]



def ob_dt(ob_list, t_list, ord=1):
    """time derivative of observable expectation 

    Args:
        ob_list (_type_): _description_
        t_list (_type_): _description_

    Returns:
        ob_dt_list: _description_
    """
    if ord == 1:
        ob_dt_list = [(ob_list[i + 1] - ob_list[i]) / (t_list[-1]/len(t_list))  for i in range(len(ob_list) - 1)]
    elif ord == 2:
        ob_dt_list = [(ob_list[i + 2] - 2*ob_list[i + 1] + ob_list[i]) / (0.5*t_list[-1]/len(t_list))  for i in range(len(ob_list) - 2)]
    return ob_dt_list



############################
# define spin Hamiltonians
############################

import numpy as np
from qiskit.quantum_info import SparsePauliOp
import random

class Nearest_Neighbour_1d:
    def __init__(self, n: int, Jx=0, Jy=0, Jz=0, hx=0, hy=0, hz=0, pbc=False, verbose=False, rand_field=[]):
        self.n = n
        self.Jx, self.Jy, self.Jz = Jx, Jy, Jz
        self.hx, self.hy, self.hz = hx, hy, hz

        self.xx_tuples = [('XX', [i, i + 1], Jx) for i in range(0, n-1)]
        self.yy_tuples = [('YY', [i, i + 1], Jy) for i in range(0, n-1)]
        self.zz_tuples = [('ZZ', [i, i + 1], Jz) for i in range(0, n-1)]

        if len(rand_field) == 0:
            self.rand_field = [0]*n
        elif len(rand_field) >= n:
            self.rand_field = rand_field[:n]
        else:
            raise ValueError(f'Length of random field should be at least {n}!')

        self.x_tuples = [('X', [i], (self.rand_field[i]+1)*hx) for i in range(0, n)] 
        self.y_tuples = [('Y', [i], (self.rand_field[i]+1)*hy) for i in range(0, n)] 
        self.z_tuples = [('Z', [i], (self.rand_field[i]+1)*hz) for i in range(0, n)] 

        if pbc: 
            self.xx_tuples.append(('XX', [n-1, 0], Jx))
            self.yy_tuples.append(('YY', [n-1, 0], Jy))
            self.zz_tuples.append(('ZZ', [n-1, 0], Jz))

        self.ham = SparsePauliOp.from_sparse_list([*self.xx_tuples, *self.yy_tuples, *self.zz_tuples, *self.x_tuples, *self.y_tuples, *self.z_tuples], num_qubits=n).simplify() 
        self.xyz_group()
        self.par_group()
        if verbose: 
            print('The Hamiltonian: \n', self.ham)
            print('The xyz grouping: \n', self.ham_xyz)
            print('The parity grouping: \n', self.ham_par)

    def xyz_group(self):
        self.x_terms = SparsePauliOp.from_sparse_list([*self.xx_tuples, *self.x_tuples], num_qubits=self.n).simplify()
        self.y_terms = SparsePauliOp.from_sparse_list([*self.yy_tuples, *self.y_tuples], num_qubits=self.n).simplify()
        self.z_terms = SparsePauliOp.from_sparse_list([*self.zz_tuples, *self.z_tuples], num_qubits=self.n).simplify()
        self.ham_xyz = [self.x_terms, self.y_terms, self.z_terms]
        self.ham_xyz = [item for item in self.ham_xyz if not np.all(abs(item.coeffs) == 0)]

    def par_group(self):
        self.even_terms = SparsePauliOp.from_sparse_list([*self.xx_tuples[::2], *self.yy_tuples[::2], *self.zz_tuples[::2], *self.x_tuples[::2], *self.y_tuples[::2], *self.z_tuples[::2]], num_qubits=self.n).simplify()
        self.odd_terms = SparsePauliOp.from_sparse_list([*self.xx_tuples[1::2], *self.yy_tuples[1::2], *self.zz_tuples[1::2], *self.x_tuples[1::2], *self.y_tuples[1::2], *self.z_tuples[1::2]], num_qubits=self.n).simplify()
        self.ham_par = [self.even_terms, self.odd_terms]

    # def lc_group(self, right, left, step):
    #     self.ham_lc = []


class Power_Law:
    def __init__(self, n: int, alpha: int, Jx=0, Jy=0, Jz=0, hx=0.0, hy=0.0, hz=0, pbc=False, verbose=False):
        self.n, self.alpha = n, alpha
        self.Jx, self.Jy, self.Jz = Jx, Jy, Jz
        self.hx, self.hy, self.hz = hx, hy, hz
        self.xx_tuples = [('XX', [i, j], Jx*abs(i-j)**(-alpha)) for i in range(0, n-1) for j in range(i+1, n)]
        self.yy_tuples = [('YY', [i, j], Jy*abs(i-j)**(-alpha)) for i in range(0, n-1) for j in range(i+1, n)]
        self.zz_tuples = [('ZZ', [i, j], Jz*abs(i-j)**(-alpha)) for i in range(0, n-1) for j in range(i+1, n)]
        self.x_tuples = [('X', [i], hx) for i in range(0, n)] 
        self.y_tuples = [('Y', [i], hy) for i in range(0, n)] 
        self.z_tuples = [('Z', [i], hz) for i in range(0, n)] 
        if pbc: 
            # self.xx_tuples.append(('XX', [n-1, 0], Jx))
            # self.yy_tuples.append(('YY', [n-1, 0], Jy))
            # self.zz_tuples.append(('ZZ', [n-1, 0], Jz))
            raise ValueError(f'PBC is not defined!')

        self.ham = SparsePauliOp.from_sparse_list([*self.xx_tuples, *self.yy_tuples, *self.zz_tuples, *self.x_tuples, *self.y_tuples, *self.z_tuples], num_qubits=n).simplify()
        if verbose: print('The Hamiltonian: \n', self.ham)
        self.xyz_group()

    def xyz_group(self):
        self.x_terms = SparsePauliOp.from_sparse_list([*self.xx_tuples, *self.x_tuples], self.n).simplify()
        self.y_terms = SparsePauliOp.from_sparse_list([*self.yy_tuples, *self.y_tuples], self.n).simplify()
        self.z_terms = SparsePauliOp.from_sparse_list([*self.zz_tuples, *self.z_tuples], self.n).simplify()
        self.ham_xyz = [self.x_terms, self.y_terms, self.z_terms]
        self.ham_xyz = [item for item in self.ham_xyz if not np.all(abs(item.coeffs) == 0)]

    # def parity_group(self):
    #     print('todo')
    #     return self.ham.to_matrix().todense()

# class TF_Ising_1d:
#     def __init__(self, n: int, J=1, h=0.2, g=0.0, pbc=False, verbose=False):
#         self.n = n
#         self.zz_tuples = [('ZZ', [i, i + 1], -J) for i in range(0, n-1)]
#         self.x_tuples = [('X', [i], -h) for i in range(0, n)] 
#         self.z_tuples = [('Z', [i], -g) for i in range(0, n)] 
#         if pbc: self.zz_tuples.append(('ZZ', [n-1, 0], -J))

#         self.ham = SparsePauliOp.from_sparse_list([*self.zz_tuples, *self.x_tuples, *self.z_tuples], num_qubits=n).simplify()
#         if verbose: print('The Hamiltonian: \n', self.ham)
#         self.parity_group()
#         self.xyz_group()

#     def parity_group(self):
#         # return self.ham.to_matrix().todense()
#         self.ham_parity = [SparsePauliOp.from_sparse_list([*self.zz_tuples[::2], *self.x_tuples[::2], *self.z_tuples[::2]], num_qubits=self.n).simplify(), SparsePauliOp.from_sparse_list([*self.zz_tuples[1::2], *self.x_tuples[1::2], *self.z_tuples[1::2]], num_qubits=self.n).simplify()]

#     def xyz_group(self):
#         self.ham_xyz = [SparsePauliOp.from_sparse_list([*self.zz_tuples, *self.z_tuples], num_qubits=self.n).simplify(), SparsePauliOp.from_sparse_list([*self.x_tuples], num_qubits=self.n)]

# class Heisenberg_1d:
#     def __init__(self, n: int, Jx=1, Jy=1, Jz=1, h=0.2, pbc=False, verbose=False):
#         self.xx_tuples = [('XX', [i, i + 1], -Jx) for i in range(0, n-1)]
#         self.yy_tuples = [('YY', [i, i + 1], -Jy) for i in range(0, n-1)]
#         self.zz_tuples = [('ZZ', [i, i + 1], -Jz) for i in range(0, n-1)]
#         self.x_tuples = [('X', [i], -h) for i in range(0, n)] 
#         if pbc: 
#             self.xx_tuples.append(('XX', [n-1, 0], -Jx))
#             self.yy_tuples.append(('YY', [n-1, 0], -Jy))
#             self.zz_tuples.append(('ZZ', [n-1, 0], -Jz))

#         self.ham = SparsePauliOp.from_sparse_list([*self.xx_tuples, *self.yy_tuples, *self.zz_tuples, *self.x_tuples], num_qubits=n)
#         if verbose: print('The Hamiltonian: \n', self.ham)

#     def parity_group(self):
#         print('todo')
#         return self.ham.to_matrix().todense()



import numpy as np
from scipy.linalg import expm
import os
# from openfermion.hamiltonians import jellium_model
from openfermion.utils import Grid, count_qubits
from openfermion.transforms import jordan_wigner, get_fermion_operator
from openfermion.linalg import get_sparse_operator

from openfermion.chem import MolecularData
from openfermionpyscf import run_pyscf
from openfermion.chem import geometry_from_pubchem

import openfermionpyscf as ofpyscf
import openfermion as of

import multiprocessing
from qiskit.quantum_info import SparsePauliOp

import matplotlib.pyplot as plt

FLOATING_POINT_PRECISION = 1e-10

def openfermion_matrix_list(qubit_operator, verbose=True):
    total_qubits = count_qubits(qubit_operator)
    matrix_list = []
    op_list = list(qubit_operator)
    if verbose: print('len(op_list): ', len(op_list))
    # print('op_list: ', op_list)
    for index, i in enumerate(op_list):
        # print(f'{index}: {i}')
        matrix_list.append(get_sparse_operator(i, total_qubits).toarray()) #changed from qubit operator and made no differnce
    # if verbose: print('len(op_list): ', len(op_list))
    return np.array(matrix_list)

    #Test -- shows ops are equivalent
def test_list_generator(openfermion_output):
    max_val = []
    of_generator = get_sparse_operator(openfermion_output).toarray()
    list_generator = sum(openfermion_matrix_list(openfermion_output))
    the_zero_op = of_generator - list_generator
    for i in range(the_zero_op.shape[0]):
        for j in range(the_zero_op.shape[0]):
            max_val.append((the_zero_op)[i][j])
    print(max(max_val))
    norm = np.linalg.norm(the_zero_op, ord=2)
    if norm < FLOATING_POINT_PRECISION:
        print("success!")
    else:
        print("failed!")
    return 0

# def jellium_hamiltonian(dimensions, length, spinless=True):
#     grid = Grid(dimensions=dimensions, length=length, scale=1.0)
#     # Get the momentum Hamiltonian.
#     momentum_hamiltonian = jellium_model(grid, spinless)
#     momentum_qubit_operator = jordan_wigner(momentum_hamiltonian)
#     momentum_qubit_operator.compress()

#     #Generate the matrix list
#     jellium_hamiltonian_list = openfermion_matrix_list(momentum_qubit_operator) #load this into simulator
#     print("Hamiltonian has dimensions: " + str(jellium_hamiltonian_list.shape))
#     #print(momentum_qubit_operator)
#     #test_list_generator(momentum_qubit_operator)
#     return jellium_hamiltonian_list

def ham_spec(hamiltonian_list):
    norms = []
    index = []
    zero_norms = 0
    for i in range(len(hamiltonian_list)):
        h = hamiltonian_list[i]
        spec = np.linalg.norm(h, ord=2)
        norms.append(spec)
        index.append(i)
        if spec == 0:
            zero_norms += 1
    print('norms: ', norms)
    norms.sort()
    print('sorted norms: ', norms)
    plt.figure(0)
    plt.plot(index, norms, 'o-', markeredgecolor='k')
    plt.xlabel("Index")
    plt.ylabel("Spectral Norm")
    plt.show()
    print("There are " + str(zero_norms) + " terms with 0 spectral norm")
    return norms

from functools import partial
def op2mat(ops, n_qubits, to_array=False):
    pool = multiprocessing.Pool(processes=multiprocessing.cpu_count())
    partial_get_sparse_operator = partial(get_sparse_operator, n_qubits=n_qubits)
    if not to_array:
        list_sparse_mat = pool.map(partial_get_sparse_operator, ops)
    
    pool.close()
    pool.join()

    return list_sparse_mat

class H2O:
    def __init__(self, verbose=False):
        # Set molecule parameters
        self.use_ActiveSpace = True
        self.d = 1.0
        self.angle = 104.5
        self.x1 = self.d*np.sin(np.pi * self.angle / 360)
        self.x2 = -self.d*np.sin(np.pi * self.angle / 360)
        self.z = self.d * np.cos(np.pi * self.angle / 360)
        self.geometry = [("H", (self.x1, 0.0, self.z)), ("O", (0.0, 0.0, 0.0)), ("H", (self.x2, 0.0, self.z))]
        self.occupied_indices = [0,1]
        self.active_indices = [2,3,4,5,6]
        self.basis = "sto-3g"
        self.multiplicity = 1
        self.charge = 0
        # Perform electronic structure calculations and
        # obtain Hamiltonian as an InteractionOperator
        self.molecule = MolecularData(self.geometry, self.basis, self.multiplicity, self.charge)
        self.molecule.load()
        if self.use_ActiveSpace:
            print('=====Using active space=====')
            self.molecular_hamiltonian = self.molecule.get_molecular_hamiltonian(
                occupied_indices = self.occupied_indices,
                active_indices = self.active_indices)
        else:
            self.molecular_hamiltonian = ofpyscf.generate_molecular_hamiltonian(
                self.geometry, self.basis, self.multiplicity, self.charge
            )
        # print('molecular_hamiltonian: ', self.molecular_hamiltonian)

        # Convert to a FermionOperator
        self.fermion = of.get_fermion_operator(self.molecular_hamiltonian)
        self.n_qubits = count_qubits(self.fermion)
        # self.fermi_terms = [get_sparse_operator(op, n_qubits=self.n_qubits) for op in self.fermion]
        self.fermi_terms = op2mat(self.fermion, self.n_qubits)
        # self.hamiltonian_matrix = [get_sparse_operator(self.fermion_hamiltonian).toarray()]
        # print(new_h2o_hamiltonian)

        # Convert to a QubitOperator
        self.jw = jordan_wigner(self.fermion)
        self.jw.compress()
        # self.qubit_terms = [get_sparse_operator(op, n_qubits=self.n_qubits) for op in self.jw]
        self.qubit_terms = op2mat(self.jw, self.n_qubits)
        self.hamiltonian_list = openfermion_matrix_list(self.jw, verbose=verbose)
        # self.qubit_ham_matrix = openfermion_matrix_list(self.qubit_ham)
        # assert np.allclose(expm(-1j * sum(self.qubit_terms)), expm(-1j * sum(self.fermi_terms)))

        if verbose:
            print(f'# fermion_ham_terms: {len(self.fermi_terms)}')
            print(f'# qubit_ham_terms: {len(self.qubit_terms)}')

def tuple2pstr(term, size):
    # print(term)
    # size = max([item[0] for item in term]+[1]) + 1
    # print('size; ', size)
    temp = ['I' for _ in range(size)]
    for item in term:
        # print(item)
        temp[item[0]] = item[1]

    return ''.join(temp)

def pauli_commutator(pauli_str1, pauli_str2):
    assert len(pauli_str1) == len(pauli_str2)
    product = []
    factor = 0
    phase = 1
    for index, pauli in enumerate(list(pauli_str1)):
        # print(index, pauli, pauli_str2[index])
        if pauli == 'I':
            product.append(pauli_str2[index])
        elif pauli == 'X':
            if pauli_str2[index] == 'X':
                product.append('I')
            elif pauli_str2[index] == 'Y':
                product.append('Z')
                factor += 1
                phase *= 1j
            elif pauli_str2[index] == 'Z':
                product.append('Y')
                factor += 1
                phase *= -1j
            else:
                product.append(pauli)
        elif pauli == 'Y':  
            if pauli_str2[index] == 'X':
                product.append('Z')
                factor += 1
                phase *= -1j
            elif pauli_str2[index] == 'Y':
                product.append('I')
            elif pauli_str2[index] == 'Z':
                product.append('X')
                factor += 1
                phase *= 1j
            else:
                product.append(pauli)
        elif pauli == 'Z':
            if pauli_str2[index] == 'X':
                product.append('Y')
                factor += 1
                phase *= 1j
            elif pauli_str2[index] == 'Y':
                product.append('X')
                factor += 1
                phase *= -1j
            elif pauli_str2[index] == 'Z':
                product.append('I')
            else:
                product.append(pauli)
        else:
            raise ValueError('Invalid Pauli string')

    return product, 2*(factor%2) * phase   

class Hydrogen_Chain:
    def __init__(self, chain_length: int, bond_length: float, verbose=False):

# def hydrogen_chain_hamiltonian(chain_length, bond_length): 
        self.chain_length = chain_length
        self.bond_length = bond_length
        self.hydrogen_geometry = []
        for i in range(self.chain_length):
            self.hydrogen_geometry.append(('H', (self.bond_length * i, 0, 0)))

        #print("Geometry in use:")
        #print(self.hydrogen_geometry)
        self.basis = 'sto-3g'
        if self.chain_length % 2 == 0:
            self.multiplicity = 1 #2ns+1
        else:
            self.multiplicity = 2

        # Set Hamiltonian parameters.
        self.active_space_start = 0
        self.active_space_stop = chain_length 

        # Set calculation parameters (to populate the molecular data class)
        self.run_scf = False #Hartree-Fock
        self.run_mp2 = False #2nd order Moller-Plesset (special case of R-S PT)
        self.run_cisd = False # Configuration interaction with single and double excitations
        self.run_ccsd = False #Coupled Cluster
        self.run_fci = True #Full configuration interaction
        self.verbose = False

        # Generate and populate instance of MolecularData.
        self.hydrogen = MolecularData(self.hydrogen_geometry, self.basis, self.multiplicity, description="hydrogen_chain_" + str(chain_length) +"_"+str(bond_length), filename="./data/hydrogen_" + str(chain_length) +"_"+str(bond_length))
        if os.path.exists(self.hydrogen.filename + '.hdf5'):
            self.hydrogen.load()
        else:
            self.hydrogen = run_pyscf(self.hydrogen, run_scf=self.run_scf, run_mp2=self.run_mp2, run_cisd=self.run_cisd, run_ccsd=self.run_ccsd, run_fci=self.run_fci, verbose=self.verbose)
            #two_body_integrals = hydrogen.two_body_integrals
            self.hydrogen.save()

        # Get the Hamiltonian in an active space.
        self.molecular_hamiltonian = self.hydrogen.get_molecular_hamiltonian(occupied_indices=range(self.active_space_start),
            active_indices=range(self.active_space_start, self.active_space_stop))

        # Map operator to fermions and qubits.
        self.fermion_hamiltonian = get_fermion_operator(self.molecular_hamiltonian)
        self.jw = jordan_wigner(self.fermion_hamiltonian)
        self.n_qubits = count_qubits(self.jw)
        self.qubit_terms = op2mat(self.jw, self.n_qubits)
        self.hamiltonian_list = openfermion_matrix_list(self.jw, verbose=self.verbose)
        self.l_terms = self.hamiltonian_list.shape[0]
        # self.ground_energy, self.ground_state = 

        self.pstrs, self.pstrs_coeff = [], []
        for term in self.jw.terms:
            # print(SparsePauliOp(tuple2pstr(term, self.n_qubits), self.jw.terms[term]))
            self.pstrs.append(tuple2pstr(term, self.n_qubits))
            self.pstrs_coeff.append(self.jw.terms[term])
        print(f'# pstr: {len(self.pstrs)}, {self.pstrs}')

        self.h_group = [self.qubit_terms[0]]
        self.pstr_group = [[{self.pstrs[0]: self.pstrs_coeff[0]}]]
        # print(self.pstrs[1:])
        # print(self.qubit_terms[0])
        for index, pstr in enumerate(self.pstrs[1:]):
            if pauli_commutator(pstr, self.pstrs[index-1])[1] == 0:
                self.h_group[-1] += self.qubit_terms[index+1]
                self.pstr_group[-1].append({pstr: self.pstrs_coeff[index+1]})
            else:
                self.h_group.append(self.qubit_terms[index+1])
                self.pstr_group.append([{pstr: self.pstrs_coeff[index+1]}])

        print(f'# groups: {len(self.h_group)}')
        print(self.pstr_group)

        if verbose:
            print(f'fermion_ham:\n {self.fermion_hamiltonian}')
            print(f'qubit_ham:\n {self.jw}')
            # print(f'hamiltonian_list: {self.hamiltonian_list}')
            print(f'L: {self.hamiltonian_list.shape}')
            print(f'grouped Hamiltonian: {self.h_group}')

    # return hydrogen_hamiltonian_list

def LiH():
    basis = 'sto-3g'
    multiplicity = 1
    # Set Hamiltonian parameters.
    active_space_start = 1
    active_space_stop = 3
    LiH_geometry = geometry_from_pubchem('LiH')
    #print(LiH_geometry)
    #diatomic_bond_length = 1.45
    #LiH_geometry = [('Li', (0., 0., 0.)), ('H', (0., 0., diatomic_bond_length))]
    #print(LiH_geometry)

    # Generate and populate instance of MolecularData.
    LiH_molecule = MolecularData(LiH_geometry, basis, multiplicity, description="1.45")
    LiH_molecule.load()

    # Get the Hamiltonian in an active space.
    LiH_molecular_hamiltonian = LiH_molecule.get_molecular_hamiltonian(
    occupied_indices=range(active_space_start),
    active_indices=range(active_space_start, active_space_stop))

    # Map operator to fermions and qubits.
    LiH_fermion_hamiltonian = get_fermion_operator(LiH_molecular_hamiltonian)
    LiH_qubit_hamiltonian = jordan_wigner(LiH_fermion_hamiltonian)
    LiH_hamiltonian_list = openfermion_matrix_list(LiH_qubit_hamiltonian)
    return LiH_hamiltonian_list


import setuptools
import subprocess
import os

package_version = (
    subprocess.run(["git", "describe", "--tags"], stdout=subprocess.PIPE)
    .stdout.decode("utf-8")
    .strip()
)
# print(package_version)

if "-" in package_version:
    # when not on tag, git describe outputs: "1.3.3-22-gdf81228"
    # pip has gotten strict with version numbers
    # so change it to: "1.3.3+22.git.gdf81228"
    # See: https://peps.python.org/pep-0440/#local-version-segments
    v,i,s = package_version.split("-")
    # package_version = v + "+" + i + ".git." + s
    package_version = v 

# v_strs = package_version.split(".")
# print(v_strs)
# package_version = ".".join(v_strs[:-1]) + '.' + str(int(v_strs[-1])+1)
assert "-" not in package_version
assert "." in package_version

assert os.path.isfile("quantum_simulation_recipe/version.py")
with open("quantum_simulation_recipe/VERSION", "w", encoding="utf-8") as fh:
    fh.write("%s\n" % package_version)

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="quantum-simulation-recipe",
    # version='0.1.6',
    version=package_version,
    # setuptools_git_versioning={
    #     "enabled": True,
    # },
    # setup_requires=["setuptools-git-versioning>=2.0,<3"],
    # author="Jue XU",
    author_email="xujue@connect.hku.hk",
    description="Recipe for quantum simulation",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Jue-Xu/quantum-simulation-recipe",
    packages=setuptools.find_packages(),
    # packages=setuptools.find_packages(include=["quantum_simulation_recipe"]),
    # package_data={"quantum_simulation_recipe": ["VERSION"]},
    # include_package_data=True,
    license="Apache-2.0",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        # "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        # "Operating System :: OS Independent",
    ],
    python_requires=">=3.10",
    entry_points={"console_scripts": ["quantum-simulation-recipe = quantum_simulation_recipe.main:main"]},
    install_requires=[
        "jax == 0.4.12",
        "qiskit >= 1.0.2",
        "openfermion >= 1.5.1",
        "openfermionpyscf >= 0.5",
        "matplotlib >= 3.8.2",
        "scipy >= 1.12.0",
        "numpy >= 1.23.5",
        "pandas >= 2.2.2",
    ],
)